//Microsoft Developer Studio generated resource script.
//
#include "resource.h"
#include "resdefs.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "windows.h"

#ifndef IDC_STATIC
#define IDC_STATIC -1
#endif

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.K.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENG)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_UK
#pragma code_page(1252)
#endif //_WIN32

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""windows.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


#ifndef _MAC
/////////////////////////////////////////////////////////////////////////////
//
// Version
//

VS_VERSION_INFO VERSIONINFO
 FILEVERSION __RCVERSION
 PRODUCTVERSION __RCVERSION
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "080904b0"
        BEGIN
            VALUE "Comments", "\0"
            VALUE "CompanyName", "TigerVNC team\0"
            VALUE "FileDescription", "TigerVNC server\0"
            VALUE "ProductName", "TigerVNC server\0"
            VALUE "FileVersion", __RCVERSIONSTR
            VALUE "InternalName", "winvnc\0"
            VALUE "LegalCopyright", "Copyright (C) 1999-2026 TigerVNC team and many others (see README.rst)\0"
            VALUE "LegalTrademarks", "TigerVNC\0"
            VALUE "OriginalFilename", "winvnc4.exe\0"
            VALUE "PrivateBuild", "\0"
            VALUE "ProductVersion", __VERSIONSTR
            VALUE "SpecialBuild", "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x809, 1200
    END
END

#endif    // !_MAC


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_ICON                ICON    DISCARDABLE     "winvnc.ico"
IDI_CONNECTED           ICON    DISCARDABLE     "connected.ico"
IDI_ICON_DISABLE        ICON    DISCARDABLE     "icon_dis.ico"
IDI_CONNECTED_DISABLE   ICON    DISCARDABLE     "connecte.ico"

/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDR_TRAY MENU DISCARDABLE 
BEGIN
    POPUP "Tray Menu"
    BEGIN
        MENUITEM "Control &panel",              ID_CONTR0L_PANEL
        MENUITEM SEPARATOR
        MENUITEM "&Options...",                 ID_OPTIONS
        MENUITEM "Add &new client...",          ID_CONNECT
        MENUITEM "&Disconnect clients",         ID_DISCONNECT
        MENUITEM "D&isable new clients",        ID_DISABLE_NEW_CLIENTS
        MENUITEM SEPARATOR
        MENUITEM "&Close VNC server",           ID_CLOSE
        MENUITEM "&About...",                   ID_ABOUT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUT DIALOG DISCARDABLE  0, 0, 300, 92
STYLE DS_MODALFRAME | DS_SETFOREGROUND | DS_CENTER | WS_POPUP | WS_CAPTION | 
    WS_SYSMENU
CAPTION "About TigerVNC server for Windows"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,245,70,47,15
    CONTROL         109,IDC_STATIC,"Static",SS_BITMAP,5,10,33,31
    LTEXT           ">appname<",IDC_DESCRIPTION,45,10,125,15
    LTEXT           ">version<",IDC_VERSION,170,10,72,15
    LTEXT           ">buildtime<",IDC_BUILDTIME,45,25,202,15
    LTEXT           ">copyright<",IDC_COPYRIGHT,45,40,256,15
    LTEXT           "See https://www.tigervnc.org for more information on VNC.",
                    IDC_STATIC,45,55,202,15
END

IDD_QUERY_CONNECT DIALOG DISCARDABLE  0, 0, 164, 93
STYLE DS_MODALFRAME | DS_SETFOREGROUND | DS_CENTER | WS_POPUP | WS_CAPTION | 
    WS_SYSMENU
CAPTION "VNC server : Accept connection?"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "&Reject",IDCANCEL,105,72,52,14
    PUSHBUTTON      "&Accept",IDOK,7,72,53,14
    RTEXT           "User:",IDC_STATIC,7,10,28,15,SS_CENTERIMAGE
    RTEXT           "Host:",IDC_STATIC,7,30,28,15,SS_CENTERIMAGE
    CTEXT           "Seconds until automatic reject:",IDC_STATIC,7,50,113,15,
                    SS_CENTERIMAGE
    LTEXT           "-",IDC_QUERY_COUNTDOWN,125,50,32,15,SS_CENTERIMAGE
    LTEXT           "-",IDC_QUERY_USER,40,10,117,15,SS_CENTERIMAGE
    LTEXT           "-",IDC_QUERY_HOST,40,30,117,15,SS_CENTERIMAGE
END

IDD_ADD_NEW_CLIENT DIALOG DISCARDABLE  0, 0, 177, 52
STYLE DS_MODALFRAME | DS_SETFOREGROUND | DS_CENTER | WS_POPUP | WS_VISIBLE | 
    WS_CAPTION | WS_SYSMENU
CAPTION "VNC server : Add new client"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_HOST,80,10,90,15,ES_AUTOHSCROLL
    DEFPUSHBUTTON   "OK",IDOK,80,31,40,14
    PUSHBUTTON      "Cancel",IDCANCEL,125,31,45,14
    CONTROL         109,IDC_STATIC,"Static",SS_BITMAP | SS_REALSIZEIMAGE,7,
                    10,33,31
    RTEXT           "Viewer:",IDC_STATIC,45,10,30,15,SS_CENTERIMAGE
END

IDD_CONTROL_PANEL DIALOG DISCARDABLE  0, 0, 267, 238
STYLE DS_MODALFRAME | DS_SETFOREGROUND | DS_CENTER | DS_CONTEXTHELP | 
    WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Control panel"
FONT 8, "MS Sans Serif"
BEGIN
    CONTROL         "List1",IDC_LIST_CONNECTIONS,"SysListView32",LVS_REPORT | 
                    LVS_SHOWSELALWAYS | LVS_ALIGNLEFT | LVS_NOSORTHEADER | 
                    WS_BORDER | WS_TABSTOP,7,25,253,76
    LTEXT           "Authorised clients list",IDC_STATIC_KLIENTS_LIST,87,7,
                    74,11,SS_CENTERIMAGE
    GROUPBOX        "Control of selected clients",
                    IDC_STATIC_SELECTED_KLIENTS,7,108,124,103
    PUSHBUTTON      "View-only",IDC_VIEW_ONLY,13,121,111,14
    PUSHBUTTON      "Full control ",IDC_FULL_CONTROL,13,145,112,14
    PUSHBUTTON      "Stop updating",IDC_STOP_UPDATE,13,167,111,14
    PUSHBUTTON      "Kill clients",IDC_KILL_SEL_CLIENT,13,190,111,14
    PUSHBUTTON      "Properties",IDC_PROPERTIES,144,121,111,14
    PUSHBUTTON      "Add new client",IDC_ADD_CLIENT,144,145,111,14
    PUSHBUTTON      "Kill all clients",IDC_KILL_ALL,144,167,111,14
    CONTROL         "Disable new clients",IDC_DISABLE_CLIENTS,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,144,191,111,13
    PUSHBUTTON      "Close",IDCANCEL,144,217,111,14
END


/////////////////////////////////////////////////////////////////////////////
//
// 24
//

#ifdef WIN64
IDR_MANIFEST            24      DISCARDABLE     "winvnc4.exe.manifest64"
#else
IDR_MANIFEST            24      DISCARDABLE     "winvnc4.exe.manifest"
#endif

/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_ABOUT, DIALOG
    BEGIN
        LEFTMARGIN, 5
        VERTGUIDE, 45
        VERTGUIDE, 170
        VERTGUIDE, 195
        VERTGUIDE, 242
        TOPMARGIN, 7
        BOTTOMMARGIN, 85
        HORZGUIDE, 10
        HORZGUIDE, 25
        HORZGUIDE, 40
        HORZGUIDE, 55
        HORZGUIDE, 70
    END

    IDD_QUERY_CONNECT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 157
        VERTGUIDE, 35
        VERTGUIDE, 40
        VERTGUIDE, 60
        VERTGUIDE, 120
        VERTGUIDE, 125
        TOPMARGIN, 7
        BOTTOMMARGIN, 86
        HORZGUIDE, 10
        HORZGUIDE, 25
        HORZGUIDE, 30
        HORZGUIDE, 45
        HORZGUIDE, 50
        HORZGUIDE, 65
    END

    IDD_ADD_NEW_CLIENT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 170
        VERTGUIDE, 45
        VERTGUIDE, 75
        VERTGUIDE, 80
        VERTGUIDE, 120
        VERTGUIDE, 125
        VERTGUIDE, 170
        TOPMARGIN, 7
        BOTTOMMARGIN, 45
        HORZGUIDE, 10
        HORZGUIDE, 25
        HORZGUIDE, 30
        HORZGUIDE, 45
    END

    IDD_CONTROL_PANEL, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 260
        VERTGUIDE, 13
        VERTGUIDE, 124
        VERTGUIDE, 144
        VERTGUIDE, 255
        TOPMARGIN, 7
        BOTTOMMARGIN, 231
        HORZGUIDE, 121
        HORZGUIDE, 135
        HORZGUIDE, 145
        HORZGUIDE, 159
        HORZGUIDE, 181
        HORZGUIDE, 191
        HORZGUIDE, 204
        HORZGUIDE, 217
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDB_BITMAP              BITMAP  DISCARDABLE     "winvnc.bmp"
#endif    // English (U.K.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

